/*
    ball-dW.cpp - Mlut and DT computations for weighted distances in 2D

    Original paper and sources:
      - E. Remy and E. Thiel. Medial Axis for Chamfer Distances: computing 
        Look-Up Tables and Neighbourhoods in 2D or 3D. 
        Pattern Recognition Letters, 23(6):649-661, April 2002.
        https://pageperso.lis-lab.fr/edouard.thiel/PRL2001/index.html

    CC BY-SA Edouard.Thiel@univ-amu.fr, Eric.remy@univ-amu.fr - 04/08/2023
*/

#include "ball-dW.hpp"

#define WDT_INFINITY  (INT_MAX/2)


void Ball_dW::add_weighting_in_g_mask (int x, int y, int w)
{
    Chamfer_weighting v;
    v.x = x;
    v.y = y;
    v.w = w;
    g_mask.push_back(v);
}


void Ball_dW::add_weighting_in_half_mask (int x, int y, int w)
{
    Chamfer_weighting v;
    v.x = x;
    v.y = y;
    v.w = w;
    half_mask.push_back(v);
}


void Ball_dW::add_chamfer_weighting (int x, int y, int w)
{
    if (! (0 <= y && y <= x && 0 < x && 0 < w)) {
        printf ("add_chamfer_weighting: (x = %d, y = %d, w = %d)\n",
            x, y, w);
        printf ("  does not respect 0 <= y <= x, 0 < x, 0 < w\n");
        return;
    }

    add_weighting_in_g_mask ( x, y, w);

    // populate half_mask with symmetries
    if (y == 0) {
        add_weighting_in_half_mask ( x, y, w);
        add_weighting_in_half_mask ( y, x, w);
    } else if (x == y) {
        add_weighting_in_half_mask ( x, y, w);
        add_weighting_in_half_mask (-x, y, w);
    } else {
        add_weighting_in_half_mask ( x, y, w);
        add_weighting_in_half_mask ( y, x, w);
        add_weighting_in_half_mask (-x, y, w);
        add_weighting_in_half_mask (-y, x, w);
    }
}


/*
 * Fast Cone Distance Transform Algorithm.
 *
 * Input:  L the side length.
 * Output: CTg the L^2 distance image to the origin.
*/

void Ball_dW::compute_CTg (int L, int *CTg)
{
    int x, y, xx, yy, i, min, val;

    CTg[0] = 0;

    for (x = 1; x < L; x++)
    for (y = 0; y <= x; y++)
    {
        min = WDT_INFINITY;

        for (i = 0; i < SIZE_INT(g_mask); i++)
        {
            xx = x - g_mask[i].x;
            yy = y - g_mask[i].y;

            if (0 <= yy && yy <= xx)
            {
                val = CTg[ yy*L + xx ] + g_mask[i].w;

                if (val < min) min = val;
            }
        }

        CTg[ y*L + x ] = min;
    }
}


/*
 * Fast Distance Transform in G(Z^2).
 *
 * Input:  L the side length,
 *         CTg the distance cone to the origin,
 *         R the radius of the ball in CTg.
 * Output: DTg the distance image to the background of the ball.
*/


void Ball_dW::compute_DTg (int L, int *CTg, int *DTg, int R)
{
    int x, y, p, xx, yy, min, val, i;

    for (y = L-1; y >= 0; y--)
    for (x = L-1; x >= y; x--)
    {
        p = y*L+x;
        if (CTg[p] > R) {  /* outside the ball : background */
            DTg[p] = 0;
        } else {
            min = WDT_INFINITY;

            for (i = 0; i < SIZE_INT(g_mask); i++)
            {
                xx = x + g_mask[i].x;
                yy = y + g_mask[i].y;

                if (xx < L)
                {
                    val = DTg[ yy*L + xx ] + g_mask[i].w;
                    if (val < min) min = val;
                }
            }

            DTg[p] = min;
        }
    }
}


/*
 * Compute the greatest verifiable radius of balls.
 *
 * Input:  L the side length.
 * Output: returns the greatest verifiable radius in the image.
*/

int Ball_dW::get_greatest_radius (int L)
{
    int i, res, xmax = 0, a = get_weight_a();

    for (i = 0; i < SIZE_INT(g_mask); i++)
    {
        if (g_mask[i].x > xmax) xmax = g_mask[i].x;
    }
    res = a * (L - xmax) -1;

    printf ("GreatestRadius for L = %d is %d\n", L, res);
    
    return res;
}


int Ball_dW::get_weight_a ()
{
    for (int i = 0; i < SIZE_INT(g_mask); i++)
    {
        if (g_mask[i].x == 1 && g_mask[i].y == 0)
            return g_mask[i].w;
    }
    return 1;
}


//----------------------------- PART USING OPENCV -----------------------------


void Ball_dW::compute_WDT (cv::Mat &img_int)
{
    CHECK_MAT_TYPE(img_int, CV_32SC1);

    // Forward scan
    for (int y = 0; y < img_int.rows; y++)
    for (int x = 0; x < img_int.cols; x++)
    {
        if (img_int.at<int>(y,x) == 0) continue;

        int min = WDT_INFINITY;
        for (int i = 0; i < SIZE_INT(half_mask); i++)
        {
            int xp = x - half_mask[i].x,
                yp = y - half_mask[i].y, vp, g;
            if (xp < 0 || yp < 0 || xp >= img_int.cols || yp >= img_int.rows)
                 vp = 0;         // exterior is considered to be 0
            else vp = img_int.at<int>(yp,xp);
            g = vp + half_mask[i].w;
            if (g < min) min = g;
        }

        img_int.at<int>(y,x) = min;
    }

    // Backward scan
    for (int y = img_int.rows-1; y >= 0; y--)
    for (int x = img_int.cols-1; x >= 0; x--)
    {
        if (img_int.at<int>(y,x) == 0) continue;

        int min = img_int.at<int>(y,x);
        for (int i = 0; i < SIZE_INT(half_mask); i++)
        {
            int xp = x + half_mask[i].x,
                yp = y + half_mask[i].y, vp, g;
            if (xp < 0 || yp < 0 || xp >= img_int.cols || yp >= img_int.rows)
                 vp = 0;         // exterior is considered to be 0
            else vp = img_int.at<int>(yp,xp);
            g = vp + half_mask[i].w;
            if (g < min) min = g;
        }

        img_int.at<int>(y,x) = min;
    }
}


void Ball_dW::compute_RWDT (cv::Mat &img_int)
{
    CHECK_MAT_TYPE(img_int, CV_32SC1);

    // Forward scan
    for (int y = 0; y < img_int.rows; y++)
    for (int x = 0; x < img_int.cols; x++)
    {
        int max = img_int.at<int>(y,x);
        for (int i = 0; i < SIZE_INT(half_mask); i++)
        {
            int xp = x - half_mask[i].x,
                yp = y - half_mask[i].y;
            if (xp < 0 || yp < 0 || xp >= img_int.cols || yp >= img_int.rows)
                continue;
            int g = img_int.at<int>(yp,xp) - half_mask[i].w;
            if (g > max) max = g;
        }

        img_int.at<int>(y,x) = max;
    }

    // Backward scan
    for (int y = img_int.rows-1; y >= 0; y--)
    for (int x = img_int.cols-1; x >= 0; x--)
    {
        int max = img_int.at<int>(y,x);
        for (int i = 0; i < SIZE_INT(half_mask); i++)
        {
            int xp = x + half_mask[i].x,
                yp = y + half_mask[i].y;
            if (xp < 0 || yp < 0 || xp >= img_int.cols || yp >= img_int.rows)
                continue;
            int g = img_int.at<int>(yp,xp) - half_mask[i].w;
            if (g > max) max = g;
        }

        img_int.at<int>(y,x) = max;
    }
}


void Ball_dW::compute_level_curves (cv::Mat &img_int)
{
    int a = get_weight_a();

    for (int y = 0; y < img_int.rows; y++)
    for (int x = 0; x < img_int.cols; x++)
        img_int.at<int>(y,x) /= a;
}


