/*
    ball-dpF0.cpp - Mlut and DT computations for the 0-Flakes distance in 2D 

    CC BY-SA Edouard.Thiel@univ-amu.fr, rita.zrour@univ-poitiers.fr - 04/08/2023
*/

#include "ball-dpF0.hpp"


/*
 * Fast Cone Distance Transform Algorithm.
 *
 * Input:  L the side length.
 * Output: CTg the L^2 distance image to the origin.
 *
 * In G(Z²) we have
 *   B_G(R) = { (x,y) \in G(Z²) : ((2x-1)² + (2y-1)²) <= R }
*/

void Ball_dpF0::compute_CTg (int L, int *CTg)
{
    int x, y;

    CTg[0] = 0;

    for (x = 1; x < L; x++)
    for (y = 0; y <= x; y++)
    {
        int xa = 2*x-1, ya = 2*y-1;
        CTg[ y*L + x ] = xa*xa + ya*ya;
    }
}


/*
 * Fast Distance Transform in Z^2/8.
 *
 * Input:  L the side length,
 *         DTg the shape (limited to Z^2/8).
 * Output: DTg the distance image to the border of the shape.
 *
 * Algorithm adapted from (Saito and Toriwaki, 1994)
 *
 * In G(Z²) we have
 *   DTg(P,X) = min { (2(xQ-xP)-1)² + (2(yQ-yP)-1)² : Q \in \bar X }
*/

void Ball_dpF0::compute_DTg (int L, int *CTg, int *DTg, int R)
{
    int x, xM, y, p, i, u, u2, propag, min, d1, d2;
 
    /* compute bound xM[ and verify that xM < L */
    for (xM = 0; xM < L; xM++) if (CTg[0*L+xM] > R) break;
    if (xM >= L) printf ("WARNING xM = %d is not < L = %d for R = %d\n", xM, L, R);
    
    /* First scan: x++, y-- */
    for (x = 0; x <= xM; x++)
    {
        i = 0; propag = 0;
        for (y = x; y >= 0; y--)
        {
            p = y*L+x;
            if (CTg[p] > R)   /* outside the ball : background */
                propag = 1;
            else if (propag)  /* inside the ball, mark to dist. (2i-1)² from bg */
              { i++; u = 2*i-1; DTg[p] = u*u; }
            else DTg[p] = -1; /* inside the ball, no distance propagated */
        }
    }

    // printf ("Saito first scan for R = %d, xM = %d:\n", R, xM);
    // print_image (L, DTg);
    
    /* Final scan: x++ */
    for (y = 0; y <= xM; y++)
    for (x = y; x <= xM; x++)
    {
        p = y*L+x;
        if (DTg[p] == 0) continue;
        
        min = 8*L*L;
        /* In the x++ scan, x+i must go up to the first pixel = 0,
          which can be in x+i=xM position. We assume that xM < L. */
        for (i = 0; i <= xM-x; i++)
        {
            d1 = DTg[p+i];
            if (d1 == -1) continue;
            // Attention: in bg, DTg[p+i] remained at 0, we must add (-1)²
            if (d1 == 0) d1 = 1;
            u = 2*i-1; u2 = u*u; d2 = d1 + u2;
            if (d2 < min) min = d2;
            else if (u2 > min) break;
        }
 
        DTg[p] = min;
    }
}


/*
 * Compute the greatest verifiable radius of balls.
 *
 * Input:  L the side length.
 * Output: returns the greatest verifiable radius in the image.
*/

int Ball_dpF0::get_greatest_radius (int L)
{
    int res = (2*L-3)*(2*L-3)-1;

    printf ("GreatestRadius for L = %d is %d\n", L, res);
    
    return res;
}


//----------------------------- PART USING OPENCV -----------------------------


int Ball_dpF0::min_column_SEDT (int col[], int y, int rows)
{
    if (col[y] == 0) return 0;

    int u, u2, d1, d2, min = 1 + col[y];

    for (int k = 1; y-k >= 0; k++) {
        u = 2*k-1; u2 = u*u; if (u2 >= min) break;
        d1 = col[y-k]; if (d1 == 0) d1 = 1;
        d2 = d1 + u2;  if (d2 < min) min = d2;
    }
    u = 2*(y+1)-1; d2 = 1 + u*u;            // exterior is considered to be 0
    if (d2 < min) min = d2;

    for (int k = 1; y+k < rows; k++) {
        u = 2*k-1; u2 = u*u; if (u2 >= min) break;
        d1 = col[y+k]; if (d1 == 0) d1 = 1;
        d2 = d1 + u2;  if (d2 < min) min = d2;
    }
    u = 2*(rows-y)-1; d2 = 1 + u*u;         // exterior is considered to be 0
    if (d2 < min) min = d2;

    return min;
}


void Ball_dpF0::compute_SEDT_Saito_Toriwaki (cv::Mat &img_int)
{
    CHECK_MAT_TYPE(img_int, CV_32SC1);

    int k, u2, u;

    // Scan of lines
    for (int y = 0; y < img_int.rows; y++)
    {
        // Forward
        k = 0;      // exterior is considered to be 0
        for (int x = 0; x < img_int.cols; x++)
        {
            if (img_int.at<int>(y,x) == 0) 
                k = 0;
            else { 
                k++; u = 2*k-1; u2 = u*u;
                img_int.at<int>(y,x) = u2; 
            }
        }

        // Backward
        k = 0;      // exterior is considered to be 0
        for (int x = img_int.cols-1 ; x >= 0; x--)
        {
            if (img_int.at<int>(y,x) == 0) 
                k = 0;
            else {
                k++; u = (2*k-1); u2 = u*u;
                if (img_int.at<int>(y,x) > u2) img_int.at<int>(y,x) = u2; 
            }
        }
    }

    // Scan of columns
    for (int x = 0; x < img_int.cols; x++)
    {
        int col[img_int.rows];
        for (int y = 0; y < img_int.rows; y++)
            col[y] = img_int.at<int>(y,x);

        for (int y = 0; y < img_int.rows; y++)
            img_int.at<int>(y,x) = min_column_SEDT (col, y, img_int.rows);
    }
}


int Ball_dpF0::max_buffer_Saito_rev (int buf[], int pos, int cmax, int scan_nb)
{
    int u, u2, d1, d2, max = abs(buf[pos]);
    
    for (int k = 1; pos-k >= 0; k++) {
        d1 = buf[pos-k]; if (d1 == 0) continue;
        u = 2*k-1; u2 = u*u;
        if (d1 > 0) d2 = d1-u2-1; else d2 = -d1-u2+1;
        if (d2 > max) max = d2;
    }

    for (int k = 1; pos+k < cmax; k++) {
        d1 = buf[pos+k]; if (d1 == 0) continue;
        u = 2*k-1; u2 = u*u;
        if (d1 > 0) d2 = d1-u2-1; else d2 = -d1-u2+1;
        if (d2 > max) max = d2;
    }

    // On first scan, generated values are marked negative
    if (scan_nb == 1 && max > buf[pos]) return -max;
    return max;
}


void Ball_dpF0::compute_RSEDT_Saito_Toriwaki (cv::Mat &img_int)
{
    CHECK_MAT_TYPE(img_int, CV_32SC1);

    int buf[std::max(img_int.cols, img_int.rows)];

    // Scan of the lines
    for (int y = 0; y < img_int.rows; y++)
    {
        // Copy line into buf
        for (int x = 0; x < img_int.cols; x++)
            buf[x] = img_int.at<int>(y,x);

        // Calculate max in buf and store in image
        for (int x = 0; x < img_int.cols; x++)
            img_int.at<int>(y,x) = max_buffer_Saito_rev (buf, x, img_int.cols, 1);
    }

    // Scan of the columns
    for (int x = 0; x < img_int.cols; x++)
    {
        // Copy line into buf
        for (int y = 0; y < img_int.rows; y++)
            buf[y] = img_int.at<int>(y,x);

        // Calculate max in buf and store in image
        for (int y = 0; y < img_int.rows; y++)
            img_int.at<int>(y,x) = max_buffer_Saito_rev (buf, y, img_int.rows, 2);
    }
}


void Ball_dpF0::compute_level_curves (cv::Mat &img_int)
{
    for (int y = 0; y < img_int.rows; y++)
    for (int x = 0; x < img_int.cols; x++)
        if (img_int.at<int>(y,x) > 0) 
            img_int.at<int>(y,x) = (sqrt(img_int.at<int>(y,x))+1)/2;
}


