/*
    ball-dpF1.cpp - Mlut and DT computations for the 1-Flakes distance in 2D 

    CC BY-SA Edouard.Thiel@univ-amu.fr, rita.zrour@univ-poitiers.fr - 04/08/2023
*/

#include "ball-dpF1.hpp"


/*
 * Fast Cone Distance Transform Algorithm.
 *
 * Input:  L the side length.
 * Output: CTg the L^2 distance image to the origin.
 *
 * In G(Z²) we have
 *   B_G(R) = { (x,y) \in G(Z²) : ((2x-1)² + (2y)²) <= R }
*/

void Ball_dpF1::compute_CTg (int L, int *CTg)
{
    int x, y;

    CTg[0] = 0;

    for (x = 1; x < L; x++)
    for (y = 0; y <= x; y++)
    {
        int xa = 2*x-1, ya = 2*y;
        CTg[ y*L + x ] = xa*xa + ya*ya;
    }
}


/*
 * Fast Distance Transform in Z^2/8.
 *
 * Input:  L the side length,
 *         DTg the shape (limited to Z^2/8).
 * Output: DTg the distance image to the border of the shape.
 *
 * Algorithm adapted from (Saito and Toriwaki, 1994)
 *
 * In G(Z²) we have
 *   DTg(P,X) = min { (2(xQ-xP)-1)² + (2(yQ-yP))² : Q \in \bar X }
*/

void Ball_dpF1::compute_DTg (int L, int *CTg, int *DTg, int R)
{
    int x, xM, y, p, i, u, u2, propag, min, d1, d2;
 
    /* compute bound xM[ and verify that xM < L */
    for (xM = 0; xM < L; xM++) if (CTg[0*L+xM] > R) break;
    if (xM >= L) printf ("WARNING xM = %d is not < L = %d for R = %d\n", xM, L, R);
    
    /* First scan: x++, y-- */
    for (x = 0; x <= xM; x++)
    {
        i = 0; propag = 0;
        for (y = x; y >= 0; y--)
        {
            p = y*L+x;
            if (CTg[p] > R)   /* outside the ball : background */
                propag = 1;
            else if (propag)  /* inside the ball, mark to dist. (2i)² from bg */
              { i++; u = 2*i; DTg[p] = u*u; }
            else DTg[p] = -1; /* inside the ball, no distance propagated */
        }
    }

    // printf ("Saito first scan for R = %d, xM = %d:\n", R, xM);
    // print_image (L, DTg);
    
    /* Final scan: x++ */
    for (y = 0; y <= xM; y++)
    for (x = y; x <= xM; x++)
    {
        p = y*L+x;
        if (DTg[p] == 0) continue;
        
        min = 8*L*L;
        /* In the x++ scan, x+i must go up to the first pixel = 0,
          which can be in x+i=xM position. We assume that xM < L. */
        for (i = 0; i <= xM-x; i++)
        {
            d1 = DTg[p+i];
            if (d1 == -1) continue;
            u = 2*i-1; u2 = u*u; d2 = d1 + u2;
            if (d2 < min) min = d2;
            else if (u2 > min) break;
        }
 
        DTg[p] = min;
    }
}


/*
 * Compute the greatest verifiable radius of balls.
 *
 * Input:  L the side length.
 * Output: returns the greatest verifiable radius in the image.
*/

int Ball_dpF1::get_greatest_radius (int L)
{
    int res = (2*L-3)*(2*L-3)-1;

    printf ("GreatestRadius for L = %d is %d\n", L, res);
    
    return res;
}


//----------------------------- PART USING OPENCV -----------------------------


// We square the column values here.
// Calculate MIN ( (2*x-1)² + (2*y)² , (2*x)² + (2*y-1)² )

int Ball_dpF1::min_column_SEDT (int col[], int y, int rows)
{

    if (col[y] == 0) return 0;
    int min = (2*col[y]-1)*(2*col[y]-1), d2, c, k;

    for (k = 1; y-k >= 0; k++) {
        if ((2*k-1)*(2*k-1) >= min) break;
        c = col[y-k];
        d2 = std::min( 
            (2*c-1)*(2*c-1) + (2*k)  *(2*k  ),
            (2*c  )*(2*c  ) + (2*k-1)*(2*k-1) );
        if (d2 < min) min = d2;
    }

    c = 0; k = y+1;
    d2 = std::min( 
        (2*c-1)*(2*c-1) + (2*k)  *(2*k  ),
        (2*c  )*(2*c  ) + (2*k-1)*(2*k-1) );
    if (d2 < min) min = d2;

    for (k = 1; y+k < rows; k++) {
        if ((2*k-1)*(2*k-1) >= min) break;
        c = col[y+k];
        d2 = std::min( 
            (2*c-1)*(2*c-1) + (2*k)  *(2*k  ),
            (2*c  )*(2*c  ) + (2*k-1)*(2*k-1) );
        if (d2 < min) min = d2;
    }

    c = 0; k = rows-y;
    d2 = std::min( 
        (2*c-1)*(2*c-1) + (2*k)  *(2*k  ),
        (2*c  )*(2*c  ) + (2*k-1)*(2*k-1) );
    if (d2 < min) min = d2;

    return min;
}


void Ball_dpF1::compute_SEDT_Saito_Toriwaki (cv::Mat &img_int)
{
    CHECK_MAT_TYPE(img_int, CV_32SC1);

    int k;

    // Scan of the lines - we dont square values
    for (int y = 0; y < img_int.rows; y++)
    {
        // Forward
        k = 0;      // exterior is considered to be 0
        for (int x = 0; x < img_int.cols; x++)
        {
            if (img_int.at<int>(y,x) == 0) 
                k = 0;
            else { 
                k++;
                img_int.at<int>(y,x) = k;
            }
        }

        // Backward
        k = 0;      // exterior is considered to be 0
        for (int x = img_int.cols-1 ; x >= 0; x--)
        {
            if (img_int.at<int>(y,x) == 0) 
                k = 0;
            else {
                k++;
                if (img_int.at<int>(y,x) > k) img_int.at<int>(y,x) = k; 
            }
        }
    }

    // Scan of the columns
    for (int x = 0; x < img_int.cols; x++)
    {
        int col[img_int.rows];
        for (int y = 0; y < img_int.rows; y++)
            col[y] = img_int.at<int>(y,x);

        for (int y = 0; y < img_int.rows; y++)
            img_int.at<int>(y,x) = min_column_SEDT (col, y, img_int.rows);
    }
}


// F1DT: a = 0 or 1

int Ball_dpF1::max_buffer_Saito_rev (int buf[], int pos, int cmax, int a)
{
    int k, h, u, max = buf[pos];
    
    for (k = 1; pos-k >= 0; k++) {
        h = buf[pos-k]; if (h == 0) continue;
        u = 2*k-a; h -= u*u; if (h > max) max = h;
    }

    for (k = 1; pos+k < cmax; k++) {
        h = buf[pos+k]; if (h == 0) continue;
        u = 2*k-a; h -= u*u; if (h > max) max = h;
    }

    return max;
}


// F1DT : we use 2 images

void Ball_dpF1::compute_RSEDT_Saito_Toriwaki (cv::Mat &img_int)
{
    CHECK_MAT_TYPE(img_int, CV_32SC1);

    int tmax = std::max(img_int.cols, img_int.rows);
    int buf[tmax], buf_0[tmax], buf_1[tmax];

    cv::Mat img_0 = img_int.clone();
    cv::Mat img_1 = img_int.clone();

    // Scan of the lines
    for (int y = 0; y < img_int.rows; y++)
    {
        // Copy line into buf
        for (int x = 0; x < img_int.cols; x++)
            buf[x] = img_int.at<int>(y,x);

        // Calculate max in buf and store in two images
        for (int x = 0; x < img_int.cols; x++) {
            img_0.at<int>(y,x) = max_buffer_Saito_rev (buf, x, img_int.cols, 0);
            img_1.at<int>(y,x) = max_buffer_Saito_rev (buf, x, img_int.cols, 1);
        }
    }

    // Scan of the columns
    for (int x = 0; x < img_int.cols; x++)
    {
        // Copy columns into buf_0 et buf_1
        // (we could avoid it because we have img_0 et img_1)
        for (int y = 0; y < img_int.rows; y++) {
            buf_0[y] = img_0.at<int>(y,x);
            buf_1[y] = img_1.at<int>(y,x);
        }

        // Calculate max in the two bufs buf and store in image
        for (int y = 0; y < img_int.rows; y++) {
            img_int.at<int>(y,x) = std::max(
                max_buffer_Saito_rev (buf_0, y, img_int.rows, 1),
                max_buffer_Saito_rev (buf_1, y, img_int.rows, 0) );
        }
    }
}


void Ball_dpF1::compute_level_curves (cv::Mat &img_int)
{
    for (int y = 0; y < img_int.rows; y++)
    for (int x = 0; x < img_int.cols; x++)
        if (img_int.at<int>(y,x) > 0) 
            img_int.at<int>(y,x) = (sqrt(img_int.at<int>(y,x))+1)/2;
}


