/*
    Computation of DT and medial axis for different kind of distances

    flakes-ma-cli.cpp - command line interface

    CC BY-SA Edouard.Thiel@univ-amu.fr - 04/09/2023
*/

#include "ball-pack.hpp"
#include <vector>


#define SHIFT_ARGS(n)  { argc -= n; argv += n; }


void print_usage (std::string prog_name) {
    std::cout << "Usage: " << prog_name << " [-d dist_name] actions\n"
        << "  dist_name:\n"
        << "    " << BallPack::get_available_dist_names() << "\n"
        << "  actions:\n"
        << "    [-show-dt] -comp-mlut L\n" 
        << "    [-i in1] [-o out2] [-size h w] [-add|-sub] [-p x y v]..."
        << " -comp-rdt\n" 
        << std::endl;
}


//---------------------------------- M A I N ----------------------------------


int main (int argc, char **argv)
{
    BallPack ball_pack;
    int opt_show_dt = 0, opt_L = 5;
    bool opt_comp_mlut = false;

    struct PixelValue { int x, y, v; };
    std::vector<PixelValue> opt_pixel_values;
    int opt_size_w = 50, opt_size_h = 40;
    std::string opt_in1  = "";
    std::string opt_out2 = "tmp1.png";
    bool opt_comp_rdt = false, opt_add = false, opt_sub = false;

    // Parse args
    std::string prog_name = argv[0]; SHIFT_ARGS (1);
    if (argc == 0) { print_usage (prog_name); exit (1); }

    while (argc > 0) {
        if (!strcmp(argv[0], "-d")) {
            if (argc < 2) break;
            ball_pack.current_dist_id = BallPack::get_dist_name_by_id(argv[1]);
            if (ball_pack.current_dist_id == BallPack::M_LAST) {
                std::cout << "Error, unknown distance name\nTry: " 
                    << BallPack::get_available_dist_names() << std::endl;
                exit (1);
            }
            SHIFT_ARGS (2);
        } else if (!strcmp(argv[0], "-comp-mlut")) {
            if (argc < 2) break;
            opt_comp_mlut = true;
            opt_L = atoi(argv[1]);
            SHIFT_ARGS (2);
        } else if (!strcmp(argv[0], "-show-dt")) {
            if (argc < 1) break;
            opt_show_dt = 1;
            SHIFT_ARGS (1);
        } else if (!strcmp(argv[0], "-size")) {
            if (argc < 3) break;
            opt_size_h = atoi(argv[1]);
            opt_size_w = atoi(argv[2]);
            SHIFT_ARGS (3); 
        } else if (!strcmp(argv[0], "-p")) {
            if (argc < 4) break;
            PixelValue p{atoi(argv[1]), atoi(argv[2]), atoi(argv[3])};
            opt_pixel_values.push_back(p);
            SHIFT_ARGS (4); 
        } else if (!strcmp(argv[0], "-i")) {
            if (argc < 2) break;
            opt_in1 = argv[1];
            SHIFT_ARGS (2);
        } else if (!strcmp(argv[0], "-o")) {
            if (argc < 2) break;
            opt_out2 = argv[1];
            SHIFT_ARGS (2);
        } else if (!strcmp(argv[0], "-add")) {
            if (argc < 1) break;
            opt_add = true;
            SHIFT_ARGS (1);
        } else if (!strcmp(argv[0], "-sub")) {
            if (argc < 1) break;
            opt_sub = true;
            SHIFT_ARGS (1);
        } else if (!strcmp(argv[0], "-comp-rdt")) {
            if (argc < 1) break;
            opt_comp_rdt = true;
            SHIFT_ARGS (1);
        } else break;
    }

    // Parsing error
    if (argc > 0) {
        std::cout << "Arguments unexpected or incomplete: ";
        for (int i = 0; i < argc; i++) 
            std::cout << " " << argv[i];
        std::cout << std::endl;
        print_usage (prog_name);
        exit (1);
    }

    // Actions
    std::cout << "Distance: " << ball_pack.get_dist_name() << std::endl;

    if (opt_comp_mlut) {
        std::cout << "Computing MLutg..." << std::endl;
        ball_pack.compute_mlutg (opt_L, opt_show_dt);
    }

    if (opt_comp_rdt) {
        cv::Mat img_src = cv::Mat(opt_size_h, opt_size_w, CV_32SC1);

        if (opt_in1 != "") {
            std::cout << "Reading image " << opt_in1 << " ..." << std::endl;
            img_src = cv::imread (opt_in1, cv::IMREAD_GRAYSCALE);
            if (img_src.empty()) {
                std::cout << "  error or incompatible image format." << std::endl;
                return -1;
            } else {
                std::cout << "  successfully read " << img_src.rows << "x" 
                  << img_src.cols << " image" << std::endl;
            }
            opt_size_h = img_src.rows;
            opt_size_w = img_src.cols;
        }

        std::cout << "Creating image " << opt_size_h << "x" << opt_size_w 
            << " ..." << std::endl;
        cv::Mat img_int = cv::Mat(opt_size_h, opt_size_w, CV_32SC1);

        if (opt_pixel_values.size() > 0)
            std::cout << "Drawing pixels:" << std::endl;
        for (unsigned int i = 0; i < opt_pixel_values.size(); i++) {
            PixelValue &p = opt_pixel_values[i];
            std::cout << "    " << p.x << " " << p.y << " " << p.v << std::endl;
            img_int.at<int>(p.y,p.x) = p.v;
        }

        std::cout << "Computing RDT ..." << std::endl;
        ball_pack.compute_RDT(img_int);

        std::cout << "Making binary result ..." << std::endl;
        for (int y = 0; y < img_int.rows; y++)
        for (int x = 0; x < img_int.cols; x++)
        {
            int a = (y < img_src.rows && x < img_src.cols) ?
                    img_src.at<uchar>(y,x) : 0;
            int b = img_int.at<int>(y,x);
            if (opt_add)
                img_int.at<int>(y,x) = (a > 0 || b > 0) ? 255 : 0;
            else if (opt_sub)
                img_int.at<int>(y,x) = (a > 0 && b == 0) ? 255 : 0;
            else
                img_int.at<int>(y,x) = (b > 0) ? 255 : 0;
        }

        std::cout << "Saving image " << opt_out2 << " ..." << std::endl;
        int res = cv::imwrite (opt_out2, img_int);
        if (res) std::cout << "  successfully saved." << std::endl;
        else std::cout << "  some error occured." << std::endl;
    }

    return 0;
}

