/*
    Computation of DT and medial axis for different kind of distances

    flakes-ma-gui.cpp - graphical user interface, using Astico2D

    CC BY-SA Edouard.Thiel@univ-amu.fr - 03/08/2023
*/

#include "astico2d.hpp"
#include "ball-pack.hpp"


//----------------------- T R A N S F O R M A T I O N S -----------------------


void filter_values (cv::Mat &img_int, int filter_value)
{
    CHECK_MAT_TYPE(img_int, CV_32SC1);

    for (int y = 0; y < img_int.rows; y++)
    for (int x = 0; x < img_int.cols; x++)
        if (img_int.at<int>(y,x) <= filter_value) 
            img_int.at<int>(y,x) = 0;
}


void compare_shapes (cv::Mat &img_orig, cv::Mat &img_ma, cv::Mat &img_int)
{
    CHECK_MAT_TYPE(img_orig, CV_32SC1);
    CHECK_MAT_TYPE(img_ma, CV_32SC1);
    CHECK_MAT_TYPE(img_int, CV_32SC1);

    for (int y = 0; y < img_int.rows; y++)
    for (int x = 0; x < img_int.cols; x++)
        if (img_ma.at<int>(y,x) > 0) 
            img_int.at<int>(y,x) = 11;
        else if (img_int.at<int>(y,x) > 0) 
            img_int.at<int>(y,x) = (img_orig.at<int>(y,x) > 0) ? 8 : 14;
        else if (img_orig.at<int>(y,x) > 0) 
            img_int.at<int>(y,x) = 12 ;
}


void detect_local_maximums_N8 (cv::Mat &img_int)
{
    CHECK_MAT_TYPE(img_int, CV_32SC1);

    int nx8[] = { 1, 1, 0, -1, -1, -1, 0, 1},
        ny8[] = { 0, 1, 1, 1, 0, -1, -1, -1};

    cv::Mat img_dt = img_int.clone();  // we need to keep a copy

    for (int y = 0; y < img_int.rows; y++)
    for (int x = 0; x < img_int.cols; x++)
    {
        int g = img_dt.at<int>(y,x);
        if (g == 0) continue;

        bool is_max = true;
        for (int i = 0; i < 8; i++)
        {
            int xp = x + nx8[i],
                yp = y + ny8[i];
            if (xp < 0 || yp < 0 || xp >= img_int.cols || yp >= img_int.rows)
                continue;
            if (g <= img_dt.at<int>(yp,xp) - 1) {
                is_max = false; break;
            }
        }
        if (!is_max) img_int.at<int>(y,x) = 0;
    }
}


void select_values (cv::Mat &img_int, cv::Mat &img_dt)
{
    for (int y = 0; y < img_int.rows; y++)
    for (int x = 0; x < img_int.cols; x++)
        if (img_int.at<int>(y,x) != 0) 
            img_int.at<int>(y,x) = img_dt.at<int>(y,x);
}


//--------------------------- A P P L I C A T I O N ---------------------------


class MyApp : public Astico2D {

  public:
    BallPack ball_pack;
    int filter_value = 0, filter_max = 10000;
    int opt_show_dt = 0, opt_L = 0;

    MyApp (int argc, char **argv) : 
        Astico2D (argc, argv),
        ball_pack()
    {
        if (!init_ok) return;  // error during initialization

        create_slider ("Filter", &filter_value, filter_max);
        rename_view_window ("View - distance "+ball_pack.get_dist_name());
    }


    void print_keyboard_keys () override
    {
        // This prints the default usage
        Astico2D::print_keyboard_keys ();

        // Indicate keyboard keys and corresponding transformations
        std::cout <<
            "  d     change the distance\n"
            "  1     show DT\n"
            "  2     show medial axis on DT\n"
            "  3     filter medial axis\n"
            "  4     show RDT on filtered medial axis\n"
            "  5     compare shapes on filtered medial axis\n"
            "  6     level curves on DT\n"
            "  7     local maximums on N8 on level curves\n"
            "  8     compare shapes using RDT on filtered local maximums\n"
            "  dD    change distance (previous/next)\n"
            "  fF    change filter on medial axis\n"
        << std::endl;
    }


    bool handle_keyboard (char key) override
    {
        switch (key) {

            // Handle the flags
            case 'd' :
                ball_pack.set_prev_dist ();
                rename_view_window ("View - distance "+ball_pack.get_dist_name());
                std::cout << "Distance : " << ball_pack.get_dist_name() << std::endl;
                break; 
            case 'D' :
                ball_pack.set_next_dist ();
                rename_view_window ("View - distance "+ball_pack.get_dist_name());
                std::cout << "Distance: " << ball_pack.get_dist_name() << std::endl;
                break; 
            case 'f' : decrease_slider ("Filter", 0); break;
            case 'F' : increase_slider ("Filter", filter_max); break;

            // Add keys for process_transformations
            case '1' :
            case '2' :
            case '3' :
            case '4' :
            case '5' :
            case '6' :
            case '7' :
            case '8' :
                // Memorize the pressed key
                key_transform = key;
                // Specify mode M_LEVELS or M_COLORS
                mode_transform = M_LEVELS;
                break;

            default : return false;  // key not handled
        }
        return true;  // key handled
    }


    void process_transformations () override
    {
        // Call here your transforms depending on key_transform and mode_transform:
        // - if mode_transform is M_COLORS, you can work directly in img_col,
        //   which has the type CV_8UC3; it will be displayed as is.
        // - if mode_transform is M_LEVELS, the starting image is the thresholded
        //   image img_int, of type CV_32SC1; at the end, to be displayed, you 
        //   need to convert the image in colors in img_col, of type CV_8UC3,
        //   for instance with the help of translate_to_vga_colors.

        switch (key_transform) {
            case '1' : {
                ball_pack.compute_DT(img_int);
                translate_to_vga_colors (img_int, img_col);
                break;
            }
            case '2' : {
                ball_pack.compute_DT(img_int);
                ball_pack.compute_MA(img_int);
                translate_to_vga_colors (img_int, img_col);
                break;
            }
            case '3' : {
                ball_pack.compute_DT(img_int);
                ball_pack.compute_MA(img_int);
                filter_values(img_int, filter_value);
                translate_to_vga_colors (img_int, img_col);
                break;
            }
            case '4' : {
                ball_pack.compute_DT(img_int);
                ball_pack.compute_MA(img_int);
                filter_values(img_int, filter_value);
                ball_pack.compute_RDT(img_int);
                translate_to_vga_colors (img_int, img_col);
                break;
            }
            case '5' : {
                cv::Mat img_orig = img_int.clone();
                ball_pack.compute_DT(img_int);
                ball_pack.compute_MA(img_int);
                filter_values(img_int, filter_value);
                cv::Mat img_ma = img_int.clone();
                ball_pack.compute_RDT(img_int);
                compare_shapes (img_orig, img_ma, img_int);
                translate_to_vga_colors (img_int, img_col);
                break;
            }
            case '6' : {
                ball_pack.compute_DT(img_int);
                ball_pack.compute_level_curves(img_int);
                translate_to_vga_colors (img_int, img_col);
                break;
            }
            case '7' : {
                ball_pack.compute_DT(img_int);
                ball_pack.compute_level_curves(img_int);
                detect_local_maximums_N8 (img_int);
                translate_to_vga_colors (img_int, img_col);
                break;
            }
            case '8' : {
                ball_pack.compute_DT(img_int);
                cv::Mat img_dt = img_int.clone();
                ball_pack.compute_level_curves(img_int);
                detect_local_maximums_N8 (img_int);
                select_values (img_int, img_dt);
                filter_values(img_int, filter_value);
                cv::Mat img_ma = img_int.clone();
                ball_pack.compute_RDT(img_int);
                compare_shapes (img_dt, img_ma, img_int);
                translate_to_vga_colors (img_int, img_col);
                break;
            }
        }
    }

}; //class MyApp


//---------------------------------- M A I N ----------------------------------


int main (int argc, char **argv)
{
    MyApp app (argc, argv);
    return app.run ();
}

